/***************************************************************************************************
*
*   $Header::                                                                                      $
*
*   PDBPlus.plw -   A replacement PDB loader for IDA Pro 4.5 that uses newer versions of DBGHELP.DLL
*                   to get all local variable names into an IDA database, in addition to standard
*                   function/data names. In future this might be extended to define IDA structures
*                   and enumerations based on the data in the .PDB file.. but I need to find time
*                   first.
*
*   NOTES           This was initially based on the PDB plugin source code in the IDA Pro 4.5 SDK,
*                   which was by Ilfak Guilfanov, and which had changes by Christopher Li. As
*                   opposed to the original version, this *requires* a recent version of DBGHELP.DLL
*                   as I didn't like working with code that had to support antiquated versions of
*                   the libraries. So, while I was formatting the source to suit my own style I
*                   scrapped the stuff that I didn't need.
*
*                   This source is formatted to for on a >100 column wide display, with tabs of 4.
*                   The project supplied is for Visual Studio .NET..
*
*   CHANGES
*
*   23/2/2003   Dean    Created
*
*   19/2/2003   Dean    Applied change found by Darko to fix local-variable naming problem.
*                       Bumped version to 1.1, and included version in short name.
*
*   23.03.2003  Darko   V1.2 - Global & local static variables
*
*   29.03.2003  Darko   V1.3 - Line numbers, data types(structures). Needs new DBGHELP (6.1+) to
*                       compile.
*
*   08.04.2003  Darko   V1.4 - support for arrays of structures
*
*   11.04.2003  Darko   V1.6 - GUI, Enums and small improvements
*
*   15.04.2003  Darko   V1.6a - fix for bug reported by Dean
*
*   19.04.2003  Darko   V1.7 - struct offsets recognizing, part of code taken from RedPlait's
*                              offsets plugin
*
*   19.04.2003  Darko   V1.7c - struct offsets recognizing as a separate process, fix in regs.cpp
*
*   21.04.2003  Darko   V1.8 - enums recognizing (ONLY for data declared as enum types)
*
*   21.04.2003  Darko   V1.8b - bug fix in function naming and undecorated symbols option
*
*   15.05.2003  Dean    V1.9  - bug fix to stop recursive behaviour when dealing with classes/structs
*                               that contain typedefs of their own type.
*
***************************************************************************************************/

#include <stdlib.h>

#include "stdafx.h"
#include <dbghelp.h>

#pragma warning( disable : 4800 )   // IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4099 )   // IDA references an internal type as both class and struct..
                                    // I don't like warnings, so I've pragma'd it away for now.
#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <auto.hpp>
#include <enum.hpp>

// We define this so we don't need to bundle Microsoft's 'cvconst.h' file.
enum SymTagEnum
{
    SymTagNull,
    SymTagExe,
    SymTagCompiland,
    SymTagCompilandDetails,
    SymTagCompilandEnv,
    SymTagFunction,
    SymTagBlock,
    SymTagData,
    SymTagAnnotation,
    SymTagLabel,
    SymTagPublicSymbol,
    SymTagUDT,
    SymTagEnum,
    SymTagFunctionType,
    SymTagPointerType,
    SymTagArrayType,
    SymTagBaseType,
    SymTagTypedef,
    SymTagBaseClass,
    SymTagFriend,
    SymTagFunctionArgType,
    SymTagFuncDebugStart,
    SymTagFuncDebugEnd,
    SymTagUsingNamespace,
    SymTagVTableShape,
    SymTagVTable,
    SymTagCustom,
    SymTagThunk,
    SymTagCustomType,
    SymTagManagedType,
    SymTagDimension,
    SymTagMax
};

enum AddStructActions
{
    E_MAKE_DATA,
    E_MAKE_ARRAY,
    E_MAKE_POINTER,
    E_MAKE_USE_EXISTING_STRUCTURE
};

#define MAX_NAME   100

// Global data
int     giGlobalFunctionCount;
int     giGlobalVariableCount;
int     giLocalVariableCount;
void*   ghProcess = (void *)0xBEEFF00D;
int     giDelta;
DWORD64 dwSymbolBase = 0;

enum    ENUMSYM_MODE
{
    ENUMERATE_GLOBALS,
    ENUMERATE_LOCALS,
    ENUMERATE_LINES,
    ENUMERATE_TYPES,
    ENUMERATE_OFFSETS_GLOBAL,
    ENUMERATE_OFFSETS_LOCAL
};

ENUMSYM_MODE    geEnumerationMode;


#define CHKBX_0001 0x0001        // First Check Box
#define CHKBX_0002 0x0002        // Second Check Box
#define CHKBX_0004 0x0004        // Third Check Box
#define CHKBX_0008 0x0008        // Fourth Check Box


// Structure for passing user selected options
typedef struct tagPASOPTION {
    short int chkbx1;
    short int chkbx2;
} PASOPTION;

enum ePDB_PLUS_OPTIONS {
    E_PDB_NOTHING   = 0,
    E_PDB_LOCAL_STATIC_VARIABLES = CHKBX_0001,
    E_PDB_LOCAL_STACK_VARIABLES  = CHKBX_0002,
    E_PDB_LOCAL_COMPLEX_ARRAY_VARIABLES = CHKBX_0004,
    E_PDB_LOCAL_ENUMS            = CHKBX_0008,

    E_PDB_GLOBAL_VARIABLES      = (CHKBX_0001 << 4),
    E_PDB_GLOBAL_COMPLEX_ARRAY_VARIABLES = (CHKBX_0002 << 4),
    E_PDB_STRUCT_OFFSETS        = (CHKBX_0004 << 4),
    E_PDB_GLOBAL_ENUMS          = (CHKBX_0008 << 4),
    E_PDB_LAST_ITEM
};

PASOPTION  gOpt_stru;
PASOPTION *gOpt_ptr = &gOpt_stru;

unsigned int gPDBOptions = 0xBF;        /* default PDB options */

/* so we do not need M$ MFC included (NB: ugly, but works) */
typedef struct
{
    unsigned short vt;
    unsigned short wReserved1;
    unsigned short wReserved2;
    unsigned short wReserved3;
    union
    {
        unsigned char        bVal;
        short                iVal;
        long                 lVal;
        long                 pad[50];   /* just in case ;) */
    };
} MY_VARIANT;

/* forward declaration for local functions */
static BOOL CALLBACK EnumerateTypesProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing );
static tid_t DumpUDTTypeIndex( DWORD dwTypeIndex, unsigned nestingLevel, unsigned offset, char *name);
static tid_t CheckIfTypeExists(PSYMBOL_INFO pSymInfo, bool bCheckIfCmplxArray, bool bCheckEnums);
static tid_t GetArrayPointerTypeIndex(DWORD typeId, unsigned nestingLevel, unsigned offset, bool bPointerCheck);
static tid_t GetSubtypes(DWORD typeId, unsigned nestingLevel, unsigned offset, WCHAR *pwszTypeName);
static int MakeAddStructMember(struc_t *sptr, const char *fieldname, ulong offset, DWORD memberOffset,
                               const typeinfo_t *mt,  ulong typeLength, int what);
static int ConvertToData(ea_t ea, ulong size, char *name, typeinfo_t  tid);
static BOOL CALLBACK EnumerateLocalSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvEA );
static BOOL CALLBACK EnumerateSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing );
static flags_t GetDataFlags(ulong size);
static tid_t IsPointerToStructure(PSYMBOL_INFO pSymInfo, unsigned nestingLevel, unsigned offset);
static void PluginError( char* pcName );

extern int Off_all_funcs(ea_t struct_addr, tid_t struct_tid, bool bEnums);
extern int Off_current_func(ea_t curr_addr, ea_t struct_addr, tid_t struct_tid, bool bEnums);
extern int Off_stack_current_func(ea_t curr_addr, ea_t struct_addr, tid_t struct_tid, bool bEnums);

#define AUTO_WAIT()  do { autoWait(); } while ( !autoIsOk() )


/**********************************************************************
* Function:     IsEnumType
* Description:  Is given data type enum ?
* Parameters:   PSYMBOL_INFO pSymInfo, unsigned nestingLevel, unsigned offset
* Returns:      tid_t or BADADDR
**********************************************************************/
static tid_t IsEnumType(PSYMBOL_INFO pSymInfo, unsigned nestingLevel, unsigned offset)
{
    tid_t    ret_val  = BADADDR;
    DWORD dk  = 0;
    DWORD dkb = 0;

    DWORD typeId = 0;
    SymGetTypeInfo( ghProcess, 0, pSymInfo->TypeIndex, TI_GET_TYPEID, &typeId);
    SymGetTypeInfo( ghProcess, 0, pSymInfo->TypeIndex, TI_GET_BASETYPE, &dk );
    SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_BASETYPE, &dkb );

    if ( dk == SymTagBlock && dkb == SymTagBlock )
    {
        /* probably enum data types ?? */
        WCHAR * pwszArrayTypeName = NULL;
        if ( SymGetTypeInfo( ghProcess, 0, pSymInfo->TypeIndex, TI_GET_SYMNAME, &pwszArrayTypeName ) )
        {
            char type_name[MAX_NAME+1];

            type_name[0] = 0;
            wcstombs(type_name, pwszArrayTypeName, MAX_NAME);
            type_name[MAX_NAME] = 0;
            LocalFree( pwszArrayTypeName );
            //msg("enum name %s\n", type_name);
            /* it is an enum type */
            ret_val= get_struc_id(type_name);
        }
    }
    if ( ret_val == BADNODE )
    {
        ret_val = BADADDR;
    }
    return ret_val;
}

/**********************************************************************
* Function:     IsPointerToStructure
* Description:  Is given data type pointer to a structure ?
* Parameters:   PSYMBOL_INFO pSymInfo, unsigned nestingLevel, unsigned offset
* Returns:      tid_t or BADADDR
**********************************************************************/
static tid_t IsPointerToStructure(PSYMBOL_INFO pSymInfo, unsigned nestingLevel, unsigned offset)
{
    tid_t    ret_val = GetArrayPointerTypeIndex(pSymInfo->TypeIndex, nestingLevel, offset, true);
    if ( ret_val == BADNODE )
    {
        ret_val = BADADDR;
    }
    return ret_val;
}

/**********************************************************************
* Function:     GetDataFlags
* Description:  Determine variable's dataflag from its size
* Parameters:   ulong size
* Returns:      flags_t
**********************************************************************/
static flags_t GetDataFlags(ulong size)
{
    flags_t     dataflag = 0;

    switch ( size )                     /* size in bytes */
    {
        case 0:
            dataflag = 0;
            break;
        case 1:
            dataflag = byteflag();
            break;
        case 2:
            dataflag = wordflag();
            break;
        case 8:
            dataflag = qwrdflag();
            break;
        default:
            if ( (size % sizeof(int)) == 0 )
            {
                dataflag = dwrdflag();
            }
            else
            {
                dataflag = byteflag();
            }
            break;
    }
    return dataflag;
}

/**********************************************************************
* Function:    ConvertToData
* Description: Convert address range to variable with given name
* Parameters: ea_t ea, int size, char * name
* Returns:    != 0 if OK
**********************************************************************/
static int ConvertToData(ea_t ea, ulong size, char *name, typeinfo_t  tid)
{
    if ( size == 0 )                    /* size should be > 0 */
    {
        return 0;
    }

    do_unknown_range(ea, size, false);
    if ( tid.tid != BADADDR )
    {
        //bool bOK =
        do_data_ex(ea, struflag(), size, tid.tid);
        //msg( "PDBPlus : struct %s %d %X %d\n", name, size, tid.tid, (ulong)bOK);
    }
    else
    {
        do_data_ex(ea, GetDataFlags(size), size, BADNODE);
    }
    return set_name(ea, name, SN_NOWARN);
}

/***************************************************************************************************
*
*   FUNCTION        EnumerateLocalSymbolsProc
*
*   DESCRIPTION     This callback is used to process symbols that exist within the scope of a
*                   particular function. In our case we're using it to look for register relative
*                   symbols, as this gets us local variables that are stored as offsets from BP.
*
*   INPUTS          pSymInfo        -       A filled in SYMBOL_INFO structure.
*
*                   ulSymbolSize    -       Size of the symbol we're processing.
*
*                   pvEA            -       This is the delta-adjusted effective address of the
*                                           function we're processing local symbols for.
*
*   RESULT          This function should return 'FALSE' if, for any reason, we need to stop the
*                   symbol enumeration. Otherwise return 'TRUE'.
*
*   NOTES           I've encountered local symbols that are not marked as register relative. These
*                   typically contain real addresses in the pSymInfo->Address field (rather than
*                   a frame offset). I believe these are static locals, which exist within the .data
*                   or .rdata segments.. in future I might see if I can create names for those
*                   variables too.
*
***************************************************************************************************/

static BOOL CALLBACK EnumerateLocalSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvEA )
{
    ea_t    ea      = ( ea_t )( pvEA );
    int     iOffset = ( int )( ( DWORD )pSymInfo->Address );
    typeinfo_t  tid;

    // We only resolve register relative offsets...
    if ( (pSymInfo->Flags & IMAGEHLP_SYMBOL_INFO_REGRELATIVE) &&
         (gPDBOptions & E_PDB_LOCAL_STACK_VARIABLES) )
    {
        EnumerateTypesProc(pSymInfo, ulSymbolSize, (PVOID) 1);

        struc_t*    pstFrame = get_frame( ea );
        if ( pstFrame == NULL )
        {
            /* something is wrong with this function -> abandon ship */
            //msg("PDBPlus : abandon ship for '%s' at %X\n", pSymInfo->Name, ea);
            return true;
        }
        int         iRealOffset = iOffset + get_func( ea )->frsize;

        del_struc_members(pstFrame, iRealOffset, iRealOffset + pSymInfo->Size );
        tid.tid = CheckIfTypeExists(pSymInfo, (gPDBOptions & E_PDB_LOCAL_COMPLEX_ARRAY_VARIABLES), false);
        if ( tid.tid != BADADDR )
        {
            add_struc_member( pstFrame, pSymInfo->Name, iRealOffset, struflag(), &tid, pSymInfo->Size );
        }
        else
        {
            if ( gPDBOptions & E_PDB_LOCAL_ENUMS )
            {
                tid.ec.tid = CheckIfTypeExists(pSymInfo, false, true);
                tid.ec.serial = 0;
            }
            if ( tid.ec.tid != BADADDR )
            {
                add_struc_member( pstFrame, pSymInfo->Name, iRealOffset,
                                  enumflag(), &tid, pSymInfo->Size );
            }
            else
            {
                add_struc_member( pstFrame, pSymInfo->Name, iRealOffset,
                                  GetDataFlags(pSymInfo->Size), NULL, pSymInfo->Size );
            }
        }
        save_struc(pstFrame);
        giLocalVariableCount++;
    }
    if ( (pSymInfo->Tag == SymTagData) && (pSymInfo->Flags == IMAGEHLP_SYMBOL_INFO_LOCAL) )
    {
        EnumerateTypesProc(pSymInfo, ulSymbolSize, (PVOID) 1);
        tid.tid = CheckIfTypeExists(pSymInfo, (gPDBOptions & E_PDB_LOCAL_COMPLEX_ARRAY_VARIABLES), false);
        giLocalVariableCount +=
            ConvertToData((ea_t)(pSymInfo->Address), pSymInfo->Size, pSymInfo->Name, tid);
    }
    return true;
}

/**********************************************************************
* Function:    EnumerateLocalOffsets
* Description: Process structure offsets for local symbols
* Parameters:  PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvEA
* Returns:     true
**********************************************************************/
static BOOL CALLBACK EnumerateLocalOffsets( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvEA )
{
    ea_t    ea      = ( ea_t )( pvEA );
    typeinfo_t  tid;

    // We only resolve register relative offsets...
    if ( (pSymInfo->Flags & IMAGEHLP_SYMBOL_INFO_REGRELATIVE) &&
         (gPDBOptions & E_PDB_LOCAL_STACK_VARIABLES) )
    {
        struc_t*    pstFrame = get_frame( ea );
        if ( pstFrame == NULL )
        {
            /* something is wrong with this function -> abandon ship */
            //msg("PDBPlus : abandon ship for '%s' at %X\n", pSymInfo->Name, ea);
            return true;
        }

        tid.tid = CheckIfTypeExists(pSymInfo, (gPDBOptions & E_PDB_LOCAL_COMPLEX_ARRAY_VARIABLES), false);
        if ( tid.tid == BADADDR )
        {
            typeinfo_t  ar;

            ar.tid = IsEnumType(pSymInfo, 0, 0);
            if ( ar.tid != BADADDR )
            {
                Off_stack_current_func(ea, (ea_t)pSymInfo->Address, ar.tid, true);
            }

            ar.tid = IsPointerToStructure(pSymInfo, 0, 0);
            if ( ar.tid != BADADDR )
            {
                Off_stack_current_func(ea, (ea_t)pSymInfo->Address, ar.tid, false);
            }
        }
        save_struc(pstFrame);
    }

    if ( (pSymInfo->Tag == SymTagData) && (pSymInfo->Flags == IMAGEHLP_SYMBOL_INFO_LOCAL) )
    {
        typeinfo_t  ar;

        ar.tid = IsEnumType(pSymInfo, 0, 0);
        if ( ar.tid != BADADDR )
        {
            Off_current_func(ea, (ea_t)(pSymInfo->Address), ar.tid, true);
        }

        ar.tid = IsPointerToStructure(pSymInfo, 0, 0);
        if ( ar.tid != BADADDR )
        {
            //msg("PDBPlus : LOCAL STATIC pointer '%s' at %X type %s address %X \n",
            //        pSymInfo->Name, ar.tid, get_struc_name(ar.tid), pSymInfo->Address);
            Off_current_func(ea, (ea_t)(pSymInfo->Address), ar.tid, false);
        }
    }
    return true;
}


/***************************************************************************************************
*
*   FUNCTION        EnumerateSymbolsProc
*
*   DESCRIPTION     This callback has, at the present time, two operating modes. When the enumeration
*                   mode variable is set to ENUMERATE_GLOBALS, then this function attempts to
*                   name items within global scope. Symbols that are tagged as functions are named
*                   and also have 'add_func' called on them. When the enumeration mode is set to
*                   ENUMERATE_LOCALS, then we look for global function symbols and use each function
*                   as the starting point for a locally scoped enumeration that identifies local
*                   symbols. This is achieved through the use of the SymSetContext() API call, prior
*                   to the secondary call to SymEnumSymbols().
*
*   INPUTS          pSymInfo        -       A filled in SYMBOL_INFO structure.
*
*                   ulSymbolSize    -       Size of the symbol we're processing.
*
*                   pvNothing       -       Not used in this callback.
*
*   RESULT          This function should return 'FALSE' if, for any reason, we need to stop the
*                   symbol enumeration. Otherwise return 'TRUE'.
*
*   NOTES           I had to separate global & local scope enumerations because the calls to
*                   'add_func()' often cause IDA to resume autoanalysis. I need to be sure that
*                   all autoanalysis-driven frame construction is completed before setting the names
*                   of local symbols, or we'd end up not setting names for all variables.
*
***************************************************************************************************/

static BOOL CALLBACK EnumerateSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing )
{
    ea_t ea     = ( ea_t )( pSymInfo->Address + giDelta );

    // We have different behaviour depending on whether we're enumerating globals or locals.
    switch ( geEnumerationMode )
    {
        case    ENUMERATE_GLOBALS:
        {
            if ( pSymInfo->Tag == SymTagFunction )
            {
                // Global functions get a name, and also have a function created for them..
                do_unknown_range(ea, pSymInfo->Size, false);  /* Darko, 24.04.2003 21:36, bug fix */
                giGlobalFunctionCount += set_name( ea, pSymInfo->Name, SN_NOWARN );;
                add_func(ea, BADADDR);
            }
            else
            {
                if ( !(gPDBOptions & E_PDB_GLOBAL_VARIABLES) )
                {
                    break;              /* skip global variables */
                }
                // Global variables get a name.. everything else is up to IDA or the user.
                EnumerateTypesProc(pSymInfo, ulSymbolSize, (PVOID) 1);
                if ( (pSymInfo->Tag == SymTagData) && (pSymInfo->Flags == 0) )
                {
                    typeinfo_t  loc_tid;
                    loc_tid.tid = CheckIfTypeExists(pSymInfo,
                                    (gPDBOptions & E_PDB_GLOBAL_COMPLEX_ARRAY_VARIABLES), false);

                    giGlobalVariableCount +=
                        ConvertToData(ea, pSymInfo->Size, pSymInfo->Name, loc_tid);
                }
                else
                {
                    giGlobalVariableCount += set_name( ea, pSymInfo->Name, SN_NOWARN );
                }
            }
            break;
        }

        case    ENUMERATE_OFFSETS_GLOBAL :
        {
            if ( pSymInfo->Tag != SymTagFunction )
            {
                if ( !(gPDBOptions & E_PDB_GLOBAL_VARIABLES) )
                {
                    break;              /* skip global variables */
                }
                if ( (pSymInfo->Tag == SymTagData) && (pSymInfo->Flags == 0) )
                {
                    typeinfo_t  ar;

                    ar.tid = IsEnumType(pSymInfo, 0, 0);
                    if ( ar.tid != BADADDR )
                    {
                        Off_all_funcs(ea, ar.tid, true);
                    }

                    ar.tid = IsPointerToStructure(pSymInfo, 0, 0);
                    if ( ar.tid != BADADDR )
                    {
                        //msg("PDBPlus : GLOBAL pointer '%s' at %X type %s address %X \n",
                        //        pSymInfo->Name, ar.tid, get_struc_name(ar.tid), ea);
                        Off_all_funcs(ea, ar.tid, false);
                    }
                }
            }
            break;
        }


        case    ENUMERATE_OFFSETS_LOCAL:
        {
            // We only go looking for locals within the context of a function..
            if ( pSymInfo->Tag == SymTagFunction )
            {
                IMAGEHLP_STACK_FRAME    obStackFrame;
                memset( &obStackFrame, 0, sizeof( obStackFrame ) );
                obStackFrame.InstructionOffset = pSymInfo->Address;
                if ( SymSetContext( ghProcess, &obStackFrame, 0 ) )
                {
                    SymEnumSymbols( ghProcess, 0, 0, EnumerateLocalOffsets, (PVOID)ea );
                }
            }
            break;
        }

        case    ENUMERATE_LOCALS:
        {
            // We only go looking for locals within the context of a function..
            if ( pSymInfo->Tag == SymTagFunction )
            {
                IMAGEHLP_STACK_FRAME    obStackFrame;
                memset( &obStackFrame, 0, sizeof( obStackFrame ) );
                obStackFrame.InstructionOffset = pSymInfo->Address;
                if ( SymSetContext( ghProcess, &obStackFrame, 0 ) )
                {
                    SymEnumSymbols( ghProcess, 0, 0, EnumerateLocalSymbolsProc, (PVOID)ea );
                }
            }
            break;
        }

        case    ENUMERATE_TYPES:
        {
            if ( pSymInfo->Tag != SymTagFunction )
            {
                EnumerateTypesProc(pSymInfo, ulSymbolSize, (PVOID) 1);
            }
        }
    }

    // Show where we are.. better than having no feedback at all.
    showAddr(ea);

    return true;
}


/**********************************************************************
* Function:     GetArrayPointerTypeIndex
* Description:  Check if given data type is a complex array
* Parameters:   DWORD typeId, unsigned nestingLevel, unsigned offset
* Returns:      tid_t or BADNODE
**********************************************************************/
static tid_t GetArrayPointerTypeIndex(DWORD typeId, unsigned nestingLevel, unsigned offset, bool bPointerCheck)
{
    typeinfo_t      mt;
    DWORD           arrayDataType = 0;

    if ( (geEnumerationMode == ENUMERATE_GLOBALS) &&
         !(gPDBOptions & E_PDB_GLOBAL_COMPLEX_ARRAY_VARIABLES) )
    {
        /* this is ugly but works (processing of complex arrays for global variables)*/
        return BADNODE;
    }

    mt.tid = BADNODE;
    if ( ! SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_TYPEID, &arrayDataType ) )
    {
        return BADNODE;
    }

    WCHAR * pwszArrayTypeName = NULL;
    if ( SymGetTypeInfo( ghProcess, 0, arrayDataType, TI_GET_SYMNAME, &pwszArrayTypeName ) )
    {
        char type_name[MAX_NAME+1];

        type_name[0] = 0;
        wcstombs(type_name, pwszArrayTypeName, MAX_NAME);
        type_name[MAX_NAME] = 0;

        mt.tid = DumpUDTTypeIndex( arrayDataType, nestingLevel, offset, type_name);
        LocalFree( pwszArrayTypeName );

        DWORD symTagArrayPointer = 0;
        SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_SYMTAG, &symTagArrayPointer );

        //msg("PDBPlus : array or pointer Type '%s' %X %d\n", type_name, mt.tid, symTagArrayPointer);

        switch ( symTagArrayPointer )
        {
            case SymTagPointerType:     /* pointer to */
                if ( bPointerCheck )
                {
                    return mt.tid;
                }
                else
                {
                    return BADNODE;
                }
                break;

            case  SymTagArrayType:      /* array of .... */
                if ( ! bPointerCheck )
                {
                    return mt.tid;
                }
                else
                {
                    return BADNODE;
                }
                break;

             default:
                break;
        }
    }
    return BADNODE;
}

/**********************************************************************
* Function:     GetSubtypes
* Description:  elements of structures, etc ..
* Parameters:   DWORD typeId, unsigned nestingLevel, unsigned offset, WCHAR *pwszTypeName
* Returns:      tid_t or BADNODE
**********************************************************************/
static tid_t GetSubtypes(DWORD typeId, unsigned nestingLevel, unsigned offset, WCHAR *pwszTypeName)
{
    #define MAX_NESTING_LEVEL   100
    typeinfo_t      mt;
    DWORD           dwChildCount;

    bool bOK = SymGetTypeInfo(ghProcess, 0, typeId, TI_GET_CHILDRENCOUNT, &dwChildCount );
    if ( (bOK && (dwChildCount > 0)) )
    {
        char type_name[MAX_NAME+1];

        if ( NULL == pwszTypeName )
        {
            return BADNODE;
        }

        type_name[0] = 0;
        wcstombs(type_name, pwszTypeName, MAX_NAME);
        type_name[MAX_NAME] = 0;

        if ( nestingLevel < MAX_NESTING_LEVEL )
        {
            //msg("PDBPlus : recursion for data type '%s'\n", type_name);
            // Recurse to get subtypes
            mt.tid = DumpUDTTypeIndex( typeId, nestingLevel+1, offset, type_name);
            return  mt.tid;
        }
        else
        {
            msg("PDBPlus : Warning - potential self-recursion for '%s'\n", type_name );
        }
    }
    return BADNODE;
}

/**********************************************************************
* Function:     MakeAddStructMember
* Description:  Use 'add_struc_member' to make an element in IDA
* Parameters:   struc_t *sptr, const char *fieldname, ulong offset, ..
* Returns:      int
**********************************************************************/
static int MakeAddStructMember(struc_t *sptr, const char *fieldname, ulong offset, DWORD memberOffset,
                               const typeinfo_t *mt,  ulong typeLength, int what)
{
    flags_t     dataflag = 0;

    switch ( what )
    {
        case E_MAKE_ARRAY:
            dataflag = struflag();
            break;

        case E_MAKE_POINTER:
            dataflag = stroffflag();
            break;

        case E_MAKE_USE_EXISTING_STRUCTURE:
            dataflag = struflag();
            break;

        default:
            dataflag = GetDataFlags(typeLength);
            break;
    }
    if ( mt )
    {
        return add_struc_member(sptr, fieldname,
                                BADADDR,
                                dataflag, mt, (ulong)typeLength);
    }
    else
    {
        return add_struc_member(sptr, fieldname,
                                (ulong)(offset + memberOffset),
                                dataflag, NULL, (ulong)typeLength);
    }
}

/**********************************************************************
* Function:    DumpUDTTypeIndex
* Description:
* Parameters: DWORD dwTypeIndex, unsigned nestingLevel, unsigned offset, char *name
* Returns:    tid_t of new IDA structure
**********************************************************************/
static tid_t DumpUDTTypeIndex( DWORD dwTypeIndex, unsigned nestingLevel, unsigned offset, char *name)
{
    #define MAX_CHILDREN_NUM    1000

    struct FINDCHILDREN : TI_FINDCHILDREN_PARAMS
    {
        ULONG   MoreChildIds[MAX_CHILDREN_NUM];
    public:
        FINDCHILDREN(){ Count = sizeof(MoreChildIds) / sizeof(MoreChildIds[0]); }
    };

    tid_t struct_tmp_id = get_struc_id(name);

    if ( BADNODE != struct_tmp_id )
    {
        //msg( "PDBPlus : struct '%s' already exists \n", name);
        return struct_tmp_id;
    }

    tid_t struct_id     = add_struc(BADADDR, name, false);
    struc_t *sptr       = get_struc(struct_id);
    enum_t  enum_id     = 0;

    //msg( "PDBPlus : add struct %s \n", name);

    DWORD childrenCount;
    bool rez = SymGetTypeInfo( ghProcess, 0, dwTypeIndex, TI_GET_CHILDRENCOUNT, &childrenCount );
    if (  rez && (childrenCount > 0) )
    {
        // assert( childrenCount < MAX_CHILDREN_NUM );

        FINDCHILDREN children;

        children.Count = childrenCount;
        children.Start= 0;

        if ( SymGetTypeInfo( ghProcess, 0, dwTypeIndex, TI_FINDCHILDREN, &children ) )
        {
            for ( unsigned i = 0; i < childrenCount; i++ )
            {
                WCHAR * pwszMemberName = NULL;
                bool bClean1 = SymGetTypeInfo( ghProcess, 0, children.ChildId[i],
                                               TI_GET_SYMNAME, &pwszMemberName );
                if ( !bClean1 )
                {
                    continue;           /* Darko, fix, 15.04.2003 20:40 */
                }

                DWORD typeId = 0;
                SymGetTypeInfo( ghProcess, 0, children.ChildId[i], TI_GET_TYPEID, &typeId);

                WCHAR * pwszTypeName = NULL;
                bool bClean2 = SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_SYMNAME, &pwszTypeName );

                DWORD memberOffset;
                SymGetTypeInfo( ghProcess, 0, children.ChildId[i], TI_GET_OFFSET, &memberOffset );

                DWORD elementCount = 0;
                SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_COUNT, &elementCount );

                ULONG64 typeLength = 0;
                SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_LENGTH, &typeLength);

                char ch_name[MAX_NAME+1];

                ch_name[0] = 0;               /* Darko, fix, 15.04.2003 20:37 */
                if ( pwszMemberName != NULL )
                {
                    wcstombs(ch_name, pwszMemberName, MAX_NAME);
                }                             /* Darko, fix, 15.04.2003 20:37 */
                ch_name[MAX_NAME] = 0;

                DWORD symTagArrayPointer = 0;
                SymGetTypeInfo( ghProcess, 0, typeId, TI_GET_SYMTAG, &symTagArrayPointer );

                typeinfo_t *mt_ptr = NULL;
                typeinfo_t mt;

                int eWhat = E_MAKE_DATA;

				if (( symTagArrayPointer == SymTagFunction ) ||
					( symTagArrayPointer == SymTagData ) ||
					( symTagArrayPointer == SymTagBaseType ) )
				{
	                mt.tid = GetSubtypes(typeId, nestingLevel, offset, pwszTypeName);
		            if ( mt.tid != BADNODE )
					{
	                    mt_ptr = &mt;
						eWhat = E_MAKE_USE_EXISTING_STRUCTURE;
					}
				}

                if ( SymTagArrayType == symTagArrayPointer )  /* is member an array ? */
                {
                    mt.tid = GetArrayPointerTypeIndex(typeId, nestingLevel+1, offset, false);
                    if ( mt.tid != BADNODE )
                    {
                        mt_ptr = &mt;
                        eWhat = E_MAKE_ARRAY;
                    }
                }
                if ( SymTagPointerType == symTagArrayPointer )  /* is member a pointer to structure ? */
                {
                    /* TO BE DONE YET */
                    //mt.path.len = 1;
                    //mt.path.delta = 0;
                    //mt.path.ids[0] = GetArrayPointerTypeIndex(typeId, nestingLevel+1, offset, true);
                    //if ( mt.path.ids[0] != BADNODE )
                    //{
                    //    mt_ptr = &mt;
                    //    eWhat = E_MAKE_POINTER;
                    //}
                }

                if ( typeLength == 0 && symTagArrayPointer == 0 )
                {
                    MY_VARIANT      val;
                    SymGetTypeInfo( ghProcess, 0, children.ChildId[i], TI_GET_VALUE, &val);

                    /* it was enum, not typedef or other structure, so delete sptr */
                    if ( i == 0 )
                    {
                        save_struc(sptr);
                        del_struc(sptr);
                        sptr = NULL;
                    }

                    /* deal with enums or not ? */
                    if ( (gPDBOptions & E_PDB_GLOBAL_ENUMS) || (gPDBOptions & E_PDB_LOCAL_ENUMS) )
                    {
                        switch ( val.vt )
                        {
                            case 2:         /* short 16 */
                                if ( i == 0 )   /* create enum name */
                                {
                                    enum_id = add_enum(BADADDR, name, wordflag());
                                }
                                add_const(enum_id, ch_name, val.iVal, DEFMASK);
                                break;
                            case 3:         /* int 32 */
                                if ( i == 0 )   /* create enum name */
                                {
                                    enum_id = add_enum(BADADDR, name, dwrdflag());
                                }
                                add_const(enum_id, ch_name, val.lVal, DEFMASK);
                                break;
                            case 17:        /* unsigned char */
                                if ( i == 0 )   /* create enum name */
                                {
                                    enum_id = add_enum(BADADDR, name, byteflag());
                                }
                                add_const(enum_id, ch_name, val.bVal, DEFMASK);
                                break;
                            default:
                                /* not supported enum types */
                                break;
                        }
                        save_enums();
                    }
                }
                else
                {
                    MakeAddStructMember(sptr, ch_name, offset,
                               memberOffset, mt_ptr,  (ulong)typeLength, eWhat);

                    save_struc(sptr);
                    //msg( "PDBPlus : adding member '%s' at offset 0x%x size %d\n",
                    //     ch_name, (ulong)(offset + memberOffset), (ulong)typeLength);
                }

                if ( bClean2 )
                {
                    LocalFree( pwszTypeName );
                }
                if ( bClean1 )
                {
                    LocalFree( pwszMemberName );
                }
            }
        }
    }
    else
    {
        //msg( "PDBPlus : exit from types %d for %s \n", rez , name);
    }
    return struct_id;
}


/**********************************************************************
* Function:    CheckIfTypeExists
* Description: Check if given name already exists in IDA names
* Parameters:  PSYMBOL_INFO pSymInfo
* Returns:     true
**********************************************************************/
static tid_t CheckIfTypeExists(PSYMBOL_INFO pSymInfo, bool bCheckIfCmplxArray, bool bCheckEnums)
{
    tid_t ret_val           = BADADDR;
    WCHAR * pwszTypeName    = NULL;

    if ( SymGetTypeInfo(ghProcess, pSymInfo->ModBase,
                        pSymInfo->TypeIndex, TI_GET_SYMNAME, &pwszTypeName )
       )
    {
        char ch_name[MAX_NAME+1];

        wcstombs(ch_name, pwszTypeName, MAX_NAME);
        ch_name[MAX_NAME] = 0;

        ret_val = get_struc_id(ch_name);
        if ( NULL == get_struc(ret_val) )
        {
           /* it is enum, not array or struct */
            if ( !bCheckEnums )
            {
                ret_val = BADADDR;
            }
            else
            {
                ret_val = get_enum(ch_name);
                //msg("PDBPlus : chk '%s' %X\n", ch_name, (ulong)ret_val);
            }
        }

        //if ( BADNODE != ret_val )
        //{
        //   //msg( "PDBPlus : basic type for data %s is %s \n", pSymInfo->Name, ch_name);
        //}
        LocalFree( pwszTypeName );
        return ret_val;
    }
    else
    {
        if ( !bCheckIfCmplxArray )          /* if test for for array of structures */
        {
            return ret_val;
        }

        ret_val = GetArrayPointerTypeIndex(pSymInfo->TypeIndex, 0, 0, false);
        if ( ret_val == BADNODE )
        {
            ret_val = BADADDR;
        }
        return ret_val;
    }
}

/**********************************************************************
* Function:    EnumerateTypesProc
* Description: Get user defined data types
* Parameters:  PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing
* Returns:     true
**********************************************************************/
static BOOL CALLBACK EnumerateTypesProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing )
{
    bool bFreeTypeName   = false;
    WCHAR * pwszTypeName = NULL;

    // Get the "real" type index for this symbol
    DWORD typeId = 0;
    SymGetTypeInfo( ghProcess, pSymInfo->ModBase, pSymInfo->TypeIndex, TI_GET_TYPEID, &typeId);

    if ( pSymInfo->Address )
    {
        if ( SymGetTypeInfo(ghProcess, pSymInfo->ModBase,
                            pSymInfo->TypeIndex, TI_GET_SYMNAME, &pwszTypeName )
           )
        {
            bFreeTypeName = true;
        }
    }

    if ( pSymInfo->Tag == SymTagUDT )   /* user defined types(structures) only */
    {
        if ( BADNODE != get_struc_id(pSymInfo->Name) )
        {
            //msg( "PDBPlus : type %s already exists \n", pSymInfo->Name);
            return true;
        }

        //msg( "PDBPlus : new type %s \n", pSymInfo->Name);
        DumpUDTTypeIndex( pSymInfo->TypeIndex, 0, 0, pSymInfo->Name);
    }

    if ( (pvNothing != 0) && bFreeTypeName )   /* user defined types(structures) only */
    {
        char ch_name[MAX_NAME+1];

        wcstombs(ch_name, pwszTypeName, MAX_NAME);
        ch_name[MAX_NAME] = 0;

        if ( BADNODE != get_struc_id(ch_name) )
        {
            //msg( "PDBPlus : type %s already exists \n", ch_name);
            LocalFree( pwszTypeName );  /* Darko, 02.04.2003 21:25 - fix memory leak */
            return true;
        }

        //msg( "PDBPlus : new type %s \n", ch_name);
        DumpUDTTypeIndex( pSymInfo->TypeIndex, 0, 0, ch_name);
    }
    if ( bFreeTypeName )
    {
        LocalFree( pwszTypeName );
    }
    return true;
}

/**********************************************************************
* Function:    EnumerateLinesProc
* Description: Puts source code file name and line number at specified address
* Parameters:  PSRCCODEINFO  pCodeInfo, PVOID pvNothing
* Returns:     true
**********************************************************************/
static BOOL CALLBACK EnumerateLinesProc( PSRCCODEINFO  pCodeInfo, PVOID pvNothing )
{
    char comment[MAXSTR+1];

    /* get file name without path */
    char *bs_ptr = strrchr(pCodeInfo->FileName, '\\');
    if ( bs_ptr == NULL )
    {
        bs_ptr = pCodeInfo->FileName;
    }
    else
    {
        bs_ptr++;
    }

    if ( strlen(bs_ptr) + 20 > MAXSTR )
    {
        return true;                    /* avoid buffer overflow */
    }
    //msg( "PDBPlus : file %s line %d  at 0x%X\n",
    //     ptr, pCodeInfo->LineNumber, pCodeInfo->Address);

    char *rpt_cmt_ptr = get_rpt_cmt((ea_t)pCodeInfo->Address);

    if ( rpt_cmt_ptr != NULL )          /* check if comment already exists */
    {
        sprintf(comment, " (file %s line %d) ", bs_ptr, pCodeInfo->LineNumber);
        if ( NULL == strstr(rpt_cmt_ptr, comment) )    /* avoid same comments */
        {
            append_cmt((ea_t)pCodeInfo->Address, comment, true);
        }
    }
    else
    {
        sprintf(comment, " (file %s line %d) ", bs_ptr, pCodeInfo->LineNumber);
        append_cmt((ea_t)pCodeInfo->Address, comment, true);
    }
    return true;
}

/***************************************************************************************************
*
*   FUNCTION        PluginError
*
*   DESCRIPTION     Displays a system error. Used when something horrible happens..
*
*   INPUTS          pcName      -   Pointer to user-supplied error text.
*
***************************************************************************************************/

static void PluginError( char* pcName )
{
    LPVOID pvMessage;

    if ( FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL,
                        GetLastError(),
                        MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                        ( LPTSTR )&pvMessage,
                        0, NULL ) )
    {
        msg( "%s: %s\n", pcName, pvMessage );
        LocalFree( pvMessage );
    }
    else
    {
        msg( "%s: Unknown error, get error message fail\n" );
    }
}

    // build the format string that creates the dialog.
const char format[] =
    "STARTITEM 0\n"
    "HELP\n"                                            // Help
    "Copy & Paste from JCR's example\n"
    "ENDHELP\n"

    "PDB Plus Configuration dialog\n"                   // Title
    "Select options for the plugin.\n\n"                // Dialog Text

//First Group
    " Local variables in functions\n"                   // Dialog Text

    // Group #1 Check Boxes NOTE: using capital "C"
    "<# static int j; gets name#"                       // hint chkbx1
    "Static local variables :C>"                        // mask 0x0001

    "<# arg_1 to arg_N get names #"                     // hint chkbx2
    "Stack variables:C>"                                // mask 0x0002

    "<# struct x[n] gets name #"                        // hint chkbx3
    "Local Complex Arrays:C>"                           // mask 0x0004

    // NOTE the additional ">"
    "<# enums loaded #"                                // hint chkbx4
    "Load Enums:C>>\n\n\n\n\n"                         // mask 0x0008

// Second Group
    " Global variables\n"                               // Dialog Text
    // Group #2 Check Boxes NOTE: using lower case "c"
    "<# static int gName; gets name #"                  // hint chkbx1
    "static global variables:c>"                        // mask 0x0001

    "<# struct gX[n] gets name #"                       // hint chkbx2
    "Global Complex Arrays:c>"                          // mask 0x0002

    "<# struct offsets & enums recognizing #"            // hint chkbx3
    "struct offsets & enums recognizing (slow!) :c>"     // mask 0x0004

    "<# Useful for C++ in release build #"      // hint chkbx4
    "Undecorated names from PDB:c>>\n\n\n\n\n"     // mask 0x0008

    ; // End Dialog Format String


/***************************************************************************************************
*
*   FUNCTION        PluginMain
*
*   DESCRIPTION     This function contains the core plugin code.
*
***************************************************************************************************/

void idaapi PluginMain( int )
{
    char*   pcInputName_Ptr;
    char    pcInputName[255];
    bool    bPdbFound = true;
    int     bOk;

    // Reset our symbol resolution counts
    giGlobalFunctionCount   = 0;
    giGlobalVariableCount   = 0;
    giLocalVariableCount    = 0;

    // Get the load base of the current PE file
    netnode penode("$ PE header");
    ea_t    dwLoadedBase = penode.altval( -2 );
    ea_t    dwLoadedSize = 0;           /* to be done */

    // Get the input file name and try to guess the PDB file locaton If failed, ask the user
    pcInputName_Ptr = get_input_file_path();
    if ( pcInputName_Ptr != NULL )
    {
        strcpy(pcInputName, pcInputName_Ptr);
    }
    else
    {
        pcInputName[0] = 0;
    }

    if ( !qfileexist( pcInputName ) )
    {
        pcInputName_Ptr = askfile_c( false, &pcInputName[0], "Please specify the input file" );
        if ( !pcInputName_Ptr )
        {
            goto done;
        }
        else
        {
            strcpy(pcInputName, pcInputName_Ptr);
        }
    }

    gOpt_ptr->chkbx2 = (gPDBOptions >> 4) & 0x0F;
    gOpt_ptr->chkbx1 = gPDBOptions & 0x0F;
    // do the magic
    bOk = AskUsingForm_c(format, &(gOpt_ptr->chkbx1), &(gOpt_ptr->chkbx2));

    if ( !bOk )
    {
        goto unload;
    }
    else
    {
        gPDBOptions = (gOpt_ptr->chkbx2 << 4) + gOpt_ptr->chkbx1;
    }

    if ( gPDBOptions & (CHKBX_0008 << 4) )
    {
        SymSetOptions(SYMOPT_LOAD_LINES | SYMOPT_UNDNAME);
    }
    else
    {
        SymSetOptions(SYMOPT_LOAD_LINES);
    }

    // Initialise
    if ( !SymInitialize( ghProcess, 0, FALSE ) )
    {
        PluginError( "SymInitialize Failed" );
        goto done;
    }

    // If I could find out how to validate the version of DBGHELP.DLL without having to manually
    // call LoadLibary() and GetProcAddress(), I'd put it here.. but for now I don't bother.
    // < Add code here.. maybe >

    // Load the symbols..
    if ( ! ( dwSymbolBase = SymLoadModule64( ghProcess, 0, &pcInputName[0], NULL, dwLoadedBase, dwLoadedSize ) ) )
    {
        PluginError( "SymLoadModule64 Failed" );
        goto cleanup;
    }

    // Obtain symbols module information
    {
        IMAGEHLP_MODULE64   obModule;
        memset(&obModule, 0, sizeof(obModule));
        obModule.SizeOfStruct = sizeof(IMAGEHLP_MODULE64);

        SymGetModuleInfo64( ghProcess, dwSymbolBase, &obModule );

        if ( obModule.SymType == SymNone )
        {
            PluginError( "SymGetModuleInfo64 : No Symbols Found");
            goto unload;
        }
    }

    // Construct a delta that might be needed..
    giDelta = ( int )( dwLoadedBase - dwSymbolBase );

    AUTO_WAIT();                        /* start only when IDA is ready */
    refresh_idaview_anyway();

    geEnumerationMode = ENUMERATE_TYPES;
    if ( !SymEnumTypes( ghProcess, ( DWORD )dwSymbolBase, EnumerateTypesProc, 0 ) )
    {
        PluginError( "SymEnumTypes Failed" );
        bPdbFound = false;
        /* goto unload; */
    }
    msg( "PDBPlus : Types loaded\n");

    AUTO_WAIT();
    refresh_idaview_anyway();

    // Enumerate symbols to construct functions
    geEnumerationMode = ENUMERATE_GLOBALS;
    if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, "", EnumerateSymbolsProc, 0 ) )
    {
        PluginError( "SymEnumSymbols(Functions) Failed" );
        goto unload;
    }
    msg( "PDBPlus : %d global function symbols loaded\n", giGlobalFunctionCount );
    msg( "PDBPlus : %d global variable symbols loaded\n", giGlobalVariableCount );

    // Wait for IDA to complete any pending autoanalysis before dealing with locals
    msg( "PDBPlus : Waiting for function analysis to complete\n" );
    AUTO_WAIT();
    refresh_idaview_anyway();
    msg( "PDBPlus : Starting local symbol analysis\n" );

    // Enumerate symbols to construct local variables
    geEnumerationMode = ENUMERATE_LOCALS;
    if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, "", EnumerateSymbolsProc, 0 ) )
    {
        PluginError( "SymEnumSymbols(Variables) Failed" );
        goto unload;
    }
    msg( "PDBPlus : %d local variable symbols loaded\n", giLocalVariableCount );
    AUTO_WAIT();
    refresh_idaview_anyway();

    if ( bPdbFound  )
    {
        geEnumerationMode = ENUMERATE_LINES;
        if ( !SymEnumLines( ghProcess, ( DWORD )dwSymbolBase, 0, 0, EnumerateLinesProc, 0 ) )
        {
            PluginError( "SymEnumLines Failed" );
            goto unload;
        }
    }

    AUTO_WAIT();
    refresh_idaview_anyway();

    if ( gPDBOptions & E_PDB_STRUCT_OFFSETS )
    {
        msg( "PDBPlus : starting struct offsets recognizing\n");
        geEnumerationMode = ENUMERATE_OFFSETS_GLOBAL;
        if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, "", EnumerateSymbolsProc, 0 ) )
        {
            PluginError( "SymEnumSymbols(offsets global) Failed" );
            goto unload;
        }
        AUTO_WAIT();
        refresh_idaview_anyway();

        geEnumerationMode = ENUMERATE_OFFSETS_LOCAL;
        if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, "", EnumerateSymbolsProc, 0 ) )
        {
            PluginError( "SymEnumSymbols(offsets local) Failed" );
            goto unload;
        }
        AUTO_WAIT();
        msg( "PDBPlus : struct offsets recognizing completed\n");
    }

    if ( !bPdbFound )
    {
        warning("PDB file for the project is not found.\n"
                "Check is .PDB file located in the same directory as:\n '%s'\n"
                "PDBPlus is NOT effective without .PDB file!!\n", pcInputName);
    }

    refresh_idaview_anyway();

unload:
    if ( !SymUnloadModule64( ghProcess, dwSymbolBase ) )
        PluginError( "SymUnloadModule64 Failed" );

cleanup:
    if ( !SymCleanup( ghProcess ) )
        PluginError( "SymCleanup Failed" );

done:
    return;
}


// -------------------------------------------------------------------------------------------------



/***************************************************************************************************
*
*   FUNCTION        PluginStartup
*
*   DESCRIPTION     IDA will call this function only once. If this function returns PLUGIN_SKIP,
*                   IDA will never load it again. If it returns PLUGIN_OK, IDA will unload the plugin
*                   but remember that the plugin agreed to work with the database. The plugin will
*                   be loaded again if the user invokes it by pressing the hotkey or selecting it
*                   from the menu. After the second load, the plugin will stay in memory.
*
*   NOTES
*
***************************************************************************************************/

int idaapi  PluginStartup(void)
{
    // We only support running our plugin when a PE file is loaded..
    if ( inf.filetype != f_PE )
        return PLUGIN_SKIP;

    return PLUGIN_OK;
}


/***************************************************************************************************
*
*   FUNCTION        PluginShutdown
*
*   DESCRIPTION     IDA will call this function when the user asks to exit. This function is *not*
*                   called in the case of emergency exits.
*
*   NOTES           Normally this function is empty..
*
***************************************************************************************************/

void idaapi PluginShutdown(void)
{
}


/***************************************************************************************************
*
*   Strings required for IDA Pro's PLUGIN descriptor block
*
***************************************************************************************************/

char    gacPluginComment[]      =   "Load global & local debug information from a PDB file";
char    gacPluginShortName[]    =   "PDB Plus V1.9";
char    gacPluginHelpText[]     =   "PDB Plus\n"
                                    "\n"
                                    "This module allows you to load debug information about function\n"
                                    "names, local variables, data names, and structures from a PDB file.\n"
                                    "\n"
                                    "The PDB file should be in the same directory as the input file\n"
                                    "\n"
                                    "Note that a recent version of DBGHELP.DLL (6.1+) is required\n";


/***************************************************************************************************
*
*   This 'PLUGIN' data block is how IDA Pro interfaces with this plugin.
*
***************************************************************************************************/

extern "C" plugin_t PLUGIN =
{
    IDP_INTERFACE_VERSION,      // Version
    PLUGIN_UNL,                 // Plugin Flags: unload the plugin immediately after run
    PluginStartup,              // Plugin Startup
    PluginShutdown,             // Plugin Shutdown.. this can be NULL
    PluginMain,                 // Plugin Main

    gacPluginComment,           // Comment about the plugin
    gacPluginHelpText,          // Multiline help about the plugin

    gacPluginShortName,         // Preferred short name of the plugin
    "",                         // Preferred hotkey to run the plugin (no hotkey specified here)
};


